/*************************************************************************/
/* Copyright (C) 2007,2008 sacamano <m.sacamano@gmail.com>		 */
/* 									 */
/* This program is free software: you can redistribute it and/or modify	 */
/* it under the terms of the GNU General Public License as published by	 */
/* the Free Software Foundation, either version 3 of the License, or	 */
/* (at your option) any later version.					 */
/* 									 */
/* This program is distributed in the hope that it will be useful,	 */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	 */
/* GNU General Public License for more details.				 */
/* 									 */
/* You should have received a copy of the GNU General Public License	 */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*************************************************************************/

#include "consonance.h"

/* Return the path of the node that matches given data */

static GtkTreePath* find_root_node(gchar *node_data,
				   GtkTreeModel *model)
{
	GtkTreePath *path = NULL;
	GtkTreeIter iter;
	gchar *data = NULL;

	if (!gtk_tree_model_get_iter_first(model, &iter))
		return NULL;

	do {
		gtk_tree_model_get(model, &iter, L_NODE_DATA, &data, -1);
		if (data && !g_ascii_strcasecmp(data, node_data)) {
			g_free(data);
			path = gtk_tree_model_get_path(model, &iter);
			break;
		}
		g_free(data);
	} while (gtk_tree_model_iter_next(model, &iter));

	return path;
}

/* Return the path of the node of r_iter's children that matches given data */

static GtkTreePath* find_child_node(gchar *node_data,
				    GtkTreeIter r_iter,
				    GtkTreeModel *model)
{
	GtkTreePath *path = NULL;
	gchar *data = NULL;
	gint i = 0;
	GtkTreeIter t_iter;

	while (gtk_tree_model_iter_nth_child(model, &t_iter, &r_iter, i++)) {
		gtk_tree_model_get(model, &t_iter, L_NODE_DATA, &data, -1);
		if (data && !g_ascii_strcasecmp(data, node_data)) {
			g_free(data);
			path = gtk_tree_model_get_path(model, &t_iter);
			break;
		}
		g_free(data);
	}

	return path;
}

/* Return node data at the given level */

static gchar* choose_node_data(gchar *location,
			       gchar *genre,
			       gchar *artist,
			       gchar *album,
			       gchar *track,
			       gint level,
			       struct con_win *cwin)
{
	gchar *node;

	node = g_slist_nth_data(cwin->cpref->library_tree_nodes, level);

	if (!g_ascii_strcasecmp(P_FOLDER_STR, node))
		return location;
	else if (!g_ascii_strcasecmp(P_GENRE_STR, node))
		return genre;
	else if (!g_ascii_strcasecmp(P_ALBUM_STR, node))
		return album;
	else if (!g_ascii_strcasecmp(P_ARTIST_STR, node))
		return artist;
	else if (!g_ascii_strcasecmp(P_TITLE_STR, node))
		return track;
	else
		return NULL;
}

/* Return node type at the given level */

static gint choose_node_type(gint level, struct con_win *cwin)
{
	gchar *node;

	node = g_slist_nth_data(cwin->cpref->library_tree_nodes, level);

	if (!g_ascii_strcasecmp(P_FOLDER_STR, node))
		return NODE_FOLDER;
	else if (!g_ascii_strcasecmp(P_BASENAME_STR, node))
		return NODE_BASENAME;
	else if (!g_ascii_strcasecmp(P_GENRE_STR, node))
		return NODE_GENRE;
	else if (!g_ascii_strcasecmp(P_ALBUM_STR, node))
		return NODE_ALBUM;
	else if (!g_ascii_strcasecmp(P_ARTIST_STR, node))
		return NODE_ARTIST;
	else if (!g_ascii_strcasecmp(P_TITLE_STR, node))
		return NODE_TRACK;
	else
		return -1;
}

/* Return node pixbuf at the given level */

static GdkPixbuf* choose_node_pixbuf(gint level, struct con_win *cwin)
{
	gchar *node;

	node = g_slist_nth_data(cwin->cpref->library_tree_nodes, level);

	if (!g_ascii_strcasecmp(P_FOLDER_STR, node))
		return cwin->pixbuf->pixbuf_dir;
	else if (!g_ascii_strcasecmp(P_BASENAME_STR, node))
		return cwin->pixbuf->pixbuf_file;
	else if (!g_ascii_strcasecmp(P_GENRE_STR, node))
		return cwin->pixbuf->pixbuf_genre;
	else if (!g_ascii_strcasecmp(P_ALBUM_STR, node))
		return cwin->pixbuf->pixbuf_album;
	else if (!g_ascii_strcasecmp(P_ARTIST_STR, node))
		return cwin->pixbuf->pixbuf_artist;
	else if (!g_ascii_strcasecmp(P_TITLE_STR, node))
		return cwin->pixbuf->pixbuf_track;
	else
		return NULL;
}

static gint add_new_location_db(gchar *location, struct con_win *cwin)
{
	gchar *query;
	gint location_id;
	struct db_result result;

	query = g_strdup_printf("INSERT INTO LOCATION (name) VALUES ('%s')",
				location);
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("SELECT id FROM LOCATION WHERE name = '%s'",
				location);
	if (exec_sqlite_query(query, cwin, &result)) {
		location_id = atoi(result.resultp[result.no_columns]);
		sqlite3_free_table(result.resultp);
	}

	return location_id;
}

static gint add_new_artist_db(gchar *artist, struct con_win *cwin)
{
	gchar *query;
	gint artist_id;
	struct db_result result;

	query = g_strdup_printf("INSERT INTO ARTIST (name) VALUES ('%s')",
				artist);
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("SELECT id FROM ARTIST WHERE name = '%s'",
				artist);
	if (exec_sqlite_query(query, cwin, &result)) {
		artist_id = atoi(result.resultp[result.no_columns]);
		sqlite3_free_table(result.resultp);
	}

	return artist_id;
}

static gint add_new_album_db(gchar *album, gint artist_id, struct con_win *cwin)
{
	gchar *query;
	gint album_id;
	struct db_result result;

	query = g_strdup_printf("INSERT INTO ALBUM (artist, name) VALUES (%d, '%s')",
				artist_id, album);
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("SELECT id FROM ALBUM WHERE name = '%s'",
				album);
	if (exec_sqlite_query(query, cwin, &result)) {
		album_id = atoi(result.resultp[result.no_columns]);
		sqlite3_free_table(result.resultp);
	}

	return album_id;
}

static gint add_new_genre_db(gchar *genre, struct con_win *cwin)
{
	gchar *query;
	gint genre_id;
	struct db_result result;

	query = g_strdup_printf("INSERT INTO GENRE (name) VALUES ('%s')",
				genre);
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("SELECT id FROM GENRE WHERE name = '%s'",
				genre);
	if (exec_sqlite_query(query, cwin, &result)) {
		genre_id = atoi(result.resultp[result.no_columns]);
		sqlite3_free_table(result.resultp);
	}

	return genre_id;
}

static gint add_new_year_db(guint year, struct con_win *cwin)
{
	gchar *query;
	gint year_id;
	struct db_result result;

	query = g_strdup_printf("INSERT INTO YEAR (year) VALUES ('%d')",
				year);
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("SELECT id FROM YEAR WHERE year = '%d'",
				year);
	if (exec_sqlite_query(query, cwin, &result)) {
		year_id = atoi(result.resultp[result.no_columns]);
		sqlite3_free_table(result.resultp);
	}

	return year_id;
}

static gint add_new_playlist_db(const gchar *playlist, struct con_win *cwin)
{
	gchar *query;
	gint playlist_id;
	struct db_result result;

	query = g_strdup_printf("INSERT INTO PLAYLIST (name) VALUES ('%s')",
				playlist);
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("SELECT id FROM PLAYLIST WHERE name = '%s'",
				playlist);
	if (exec_sqlite_query(query, cwin, &result)) {
		playlist_id = atoi(result.resultp[result.no_columns]);
		sqlite3_free_table(result.resultp);
	}

	return playlist_id;
}

static void add_track_playlist_db(gchar *file, gint playlist_id, struct con_win *cwin)
{
	gchar *query;

	query = g_strdup_printf("INSERT INTO PLAYLIST_TRACKS (file, playlist) VALUES ('%s', %d);",
				file,
				playlist_id);
	exec_sqlite_query(query, cwin, NULL);
}

static gboolean overwrite_existing_playlist(const gchar *playlist, struct con_win *cwin)
{
	gboolean choice = FALSE;
	gint ret;
	GtkWidget *dialog;

	dialog = gtk_message_dialog_new(GTK_WINDOW(cwin->mainwindow),
					GTK_DIALOG_MODAL,
					GTK_MESSAGE_QUESTION,
					GTK_BUTTONS_YES_NO,
					"Do you want to overwrite the playlist: %s ?",
					playlist);

	ret = gtk_dialog_run(GTK_DIALOG(dialog));

	switch(ret) {
	case GTK_RESPONSE_YES:
		choice = TRUE;
		break;
	case GTK_RESPONSE_NO:
		choice = FALSE;
		break;
	default:
		break;
	}

	gtk_widget_destroy(dialog);
	return choice;
}

static void add_new_track_db(gint location_id,
			     gint artist_id,
			     gint album_id,
			     gint genre_id,
			     gint year_id,
			     guint track_no,
			     gint length,
			     gint channels,
			     gint bitrate,
			     gint samplerate,
			     gint file_type,
			     gchar *title,
			     struct con_win *cwin)
{
	gchar *query;

	query = g_strdup_printf("INSERT INTO TRACK ( \
location,   \
track_no,   \
artist,     \
album,      \
genre,      \
year,       \
bitrate,    \
samplerate, \
length,     \
channels,   \
file_type,  \
title)      \
	    \
VALUES      \
	    \
('%d', '%d', '%d', '%d', '%d', '%d', '%d', '%d', '%d', %d, '%d', '%s')",

				location_id,
				track_no,
				artist_id,
				album_id,
				genre_id,
				year_id,
				bitrate,
				samplerate,
				length,
				channels,
				file_type,
				title);
	exec_sqlite_query(query, cwin, NULL);
}

static void add_entry_db(gchar *file, struct con_win *cwin)
{
	struct musicobject *mobj;
	gchar *sfile, *stitle, *sartist, *salbum, *sgenre;
	gint location_id, artist_id, album_id, genre_id, year_id;

	mobj = new_musicobject_from_file(file);
	if (mobj) {
		sfile = sanitize_string_sqlite3(file);
		stitle = sanitize_string_sqlite3(mobj->tags->title);
		sartist = sanitize_string_sqlite3(mobj->tags->artist);
		salbum = sanitize_string_sqlite3(mobj->tags->album);
		sgenre = sanitize_string_sqlite3(mobj->tags->genre);

		/* Write location */

		location_id = add_new_location_db(sfile, cwin);

		/* Write artist */

		artist_id = add_new_artist_db(sartist, cwin);

		/* Write album */

		album_id = add_new_album_db(salbum, artist_id, cwin);

		/* Write genre */

		genre_id = add_new_genre_db(sgenre, cwin);

		/* Write year */

		year_id = add_new_year_db(mobj->tags->year, cwin);

		/* Write track */

		add_new_track_db(location_id,
				 artist_id,
				 album_id,
				 genre_id,
				 year_id,
				 mobj->tags->track_no,
				 mobj->tags->length,
				 mobj->tags->channels,
				 mobj->tags->bitrate,
				 mobj->tags->samplerate,
				 mobj->file_type,
				 stitle,
				 cwin);

		g_free(sfile);
		g_free(stitle);
		g_free(sartist);
		g_free(salbum);
		g_free(sgenre);

		delete_musicobject(mobj);
	}
}

/* Add a new entry to the library tree */

static void add_entry_library(gint location_id,
			      gchar *location,
			      gchar *genre,
			      gchar *artist,
			      gchar *album,
			      gchar *track,
			      GtkTreeModel *model,
			      struct con_win *cwin)
{
	gchar *node_data;
	gint node_type = 0, node_level = 0, tot_levels = 0;
	GdkPixbuf *node_pixbuf;
	GtkTreeIter iter = {0}, p_iter = {0};
	GtkTreePath *p_path = NULL;

	tot_levels = g_slist_length(cwin->cpref->library_tree_nodes);

	node_data = choose_node_data(location, genre, artist, album, track, node_level, cwin);
	node_type = choose_node_type(node_level, cwin);
	node_pixbuf = choose_node_pixbuf(node_level, cwin);

	/* Only root node can be of type NODE_FOLDER */

	if (node_type == NODE_FOLDER)
		node_data = get_containing_folder(location);

	/* Check if root node is already present */

	p_path = find_root_node(node_data, model);
	if (!p_path) {
		gtk_tree_store_append(GTK_TREE_STORE(model),
				      &iter,
				      NULL);
		gtk_tree_store_set(GTK_TREE_STORE(model), &iter,
				   L_PIXBUF, node_pixbuf,
				   L_NODE_DATA, node_data,
				   L_NODE_TYPE, node_type,
				   L_LOCATION_ID, 0,
				   L_VISIBILE, TRUE,
				   -1);
		p_iter = iter;
	}
	else {
		gtk_tree_model_get_iter(model, &p_iter, p_path);
		gtk_tree_path_free(p_path);
		p_path = NULL;
	}

	if (node_type == NODE_FOLDER)
		g_free(node_data);

	node_level++;

	/* Run through the remaining child nodes */

	while (node_level < tot_levels) {
		node_data = choose_node_data(location, genre, artist, album, track, node_level, cwin);
		node_type = choose_node_type(node_level, cwin);
		node_pixbuf = choose_node_pixbuf(node_level, cwin);

		if (node_type == NODE_BASENAME)
			node_data = g_path_get_basename(location);

		/* Check if the child node is already present */
		/* For leaves (title/basename), append the entry irrespective of duplicate presence */

		if ((node_type == NODE_BASENAME) || (node_type == NODE_TRACK))
			p_path = NULL;
		else
			p_path = find_child_node(node_data, p_iter, model);

		if (!p_path) {
			gtk_tree_store_append(GTK_TREE_STORE(model),
					      &iter,
					      &p_iter);
			gtk_tree_store_set(GTK_TREE_STORE(model), &iter,
					   L_PIXBUF, node_pixbuf,
					   L_NODE_DATA, node_data,
					   L_NODE_TYPE, node_type,
					   L_LOCATION_ID,
					   ((node_type == NODE_TRACK) || (node_type == NODE_BASENAME)) ?
					   location_id : 0,
					   L_VISIBILE, TRUE,
					   -1);
			p_iter = iter;
		}
		else {
			gtk_tree_model_get_iter(model, &p_iter, p_path);
			gtk_tree_path_free(p_path);
			p_path = NULL;
		}

		if (node_type == NODE_BASENAME)
			g_free(node_data);

		node_level++;
	}
}

/* Append to the given array the location ids of all the nodes under the given path */

static void get_location_ids(GtkTreePath *path,
			     GArray *loc_arr,
			     GtkTreeModel *model,
			     struct con_win *cwin)
{
	GtkTreeIter t_iter, r_iter;
	gint node_type, location_id;
	gint j = 0;

	gtk_tree_model_get_iter(model, &r_iter, path);

	/* If this path is a track, just append it to the array */

	gtk_tree_model_get(model, &r_iter, L_NODE_TYPE, &node_type, -1);
	if ((node_type == NODE_TRACK) || (node_type == NODE_BASENAME)) {
		gtk_tree_model_get(model, &r_iter, L_LOCATION_ID, &location_id, -1);
		g_array_append_val(loc_arr, location_id);
	}

	/* For all other node types do a recursive add */

	while (gtk_tree_model_iter_nth_child(model, &t_iter, &r_iter, j++)) {
		gtk_tree_model_get(model, &t_iter, L_NODE_TYPE, &node_type, -1);
		if ((node_type == NODE_TRACK) || (node_type == NODE_BASENAME)) {
			gtk_tree_model_get(model, &t_iter, L_LOCATION_ID, &location_id, -1);
			g_array_append_val(loc_arr, location_id);
		}
		else {
			path = gtk_tree_model_get_path(model, &t_iter);
			get_location_ids(path, loc_arr, model, cwin);
			gtk_tree_path_free(path);
		}
	}
}

/* Add all the tracks under the given path to the current playlist */

static void add_row_current_playlist(GtkTreePath *path, GtkTreeModel *model, struct con_win *cwin)
{
	GtkTreeIter t_iter, r_iter;
	gint node_type, location_id;
	struct musicobject *mobj = NULL;
	gint j = 0;

	/* If this path is a track, just append it to the current playlist */

	gtk_tree_model_get_iter(model, &r_iter, path);
	gtk_tree_model_get(model, &r_iter, L_NODE_TYPE, &node_type, -1);
	if ((node_type == NODE_TRACK) || (node_type == NODE_BASENAME)) {
		gtk_tree_model_get(model, &r_iter, L_LOCATION_ID, &location_id, -1);
		mobj = new_musicobject_from_db(location_id, cwin);
		if (!mobj)
			g_warning("(%s): Unable to retrieve details for location_id : %d",
				  __func__,
				  location_id);
		else
			append_current_playlist(mobj, cwin);
	}

	/* For all other node types do a recursive add */

	while (gtk_tree_model_iter_nth_child(model, &t_iter, &r_iter, j++)) {
		gtk_tree_model_get(model, &t_iter, L_NODE_TYPE, &node_type, -1);
		if ((node_type == NODE_TRACK) || (node_type == NODE_BASENAME)) {
			gtk_tree_model_get(model, &t_iter, L_LOCATION_ID, &location_id, -1);
			mobj = new_musicobject_from_db(location_id, cwin);
			if (!mobj)
				g_warning("(%s): Unable to retrieve details for location_id : %d",
					  __func__,
					  location_id);
			else
				append_current_playlist(mobj, cwin);
		}
		else {
			path = gtk_tree_model_get_path(model, &t_iter);
			add_row_current_playlist(path, model, cwin);
			gtk_tree_path_free(path);
		}
	}
}

/* Save tracks to a playlist using the given type */

static void do_save_playlist(gint playlist_id, gint type, struct con_win *cwin)
{
	gchar *file;
	GtkTreeModel *model;
	GtkTreePath *path;
	GtkTreeSelection *selection;
	GtkTreeIter iter;
	struct musicobject *mobj = NULL;
	GList *list, *i;

	model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->current_playlist));

	switch(type) {
	case SAVE_COMPLETE:
		gtk_tree_model_get_iter_first(model, &iter);
		do {
			gtk_tree_model_get(model, &iter, P_MOBJ_PTR, &mobj, -1);
			if (mobj) {
				file = sanitize_string_sqlite3(mobj->file);
				add_track_playlist_db(file, playlist_id, cwin);
				g_free(file);
			}
		} while(gtk_tree_model_iter_next(model, &iter));
		break;
	case SAVE_SELECTED:
		selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cwin->current_playlist));
		list = gtk_tree_selection_get_selected_rows(selection, NULL);

		if (list) {

			/* Add all the rows to the playlist */

			for (i=list; i != NULL; i = i->next) {
				path = i->data;
				gtk_tree_model_get_iter(model, &iter, path);
				gtk_tree_model_get(model, &iter, P_MOBJ_PTR, &mobj, -1);

				if (mobj) {
					file = sanitize_string_sqlite3(mobj->file);
					add_track_playlist_db(file, playlist_id, cwin);
					g_free(file);
				}

				gtk_tree_path_free(path);
			}
			g_list_free(list);
		}
		break;
	default:
		break;
	}
}

void flush_db(struct con_win *cwin)
{
	gchar *query;

	query = g_strdup_printf("DELETE FROM TRACK");
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("DELETE FROM LOCATION");
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("DELETE FROM ARTIST");
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("DELETE FROM ALBUM");
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("DELETE FROM GENRE");
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("DELETE FROM YEAR");
	exec_sqlite_query(query, cwin, NULL);
}

gint init_dbase_schema(struct con_win *cwin)
{
	gchar *query;

	/* Create 'TRACKS' table */

	query = g_strdup_printf("CREATE TABLE IF NOT EXISTS TRACK (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s);",
				"location INT PRIMARY KEY",
				"track_no INT",
				"artist INT",
				"album INT",
				"genre INT",
				"year INT",
				"bitrate INT",
				"length INT",
				"channels INT",
				"samplerate INT",
				"file_type INT",
				"title VARCHAR(255)");
	if (!exec_sqlite_query(query, cwin, NULL))
		return -1;

	/* Create 'LOCATION' table */

	query = g_strdup_printf("CREATE TABLE IF NOT EXISTS LOCATION (%s, %s, UNIQUE(name));",
				"id INTEGER PRIMARY KEY",
				"name TEXT");
	if (!exec_sqlite_query(query, cwin, NULL))
		return -1;


	/* Create 'ARTIST' table */

	query = g_strdup_printf("CREATE TABLE IF NOT EXISTS ARTIST (%s, %s, UNIQUE(name));",
				"id INTEGER PRIMARY KEY",
				"name VARCHAR(255)");
	if (!exec_sqlite_query(query,cwin,  NULL))
		return -1;


	/* Create 'ALBUM' table */

	query = g_strdup_printf("CREATE TABLE IF NOT EXISTS ALBUM (%s, %s, %s, UNIQUE(name));",
				"id INTEGER PRIMARY KEY",
				"artist INT",
				"name VARCHAR(255)");
	if (!exec_sqlite_query(query, cwin, NULL))
		return -1;

	/* Create 'GENRE' table */

	query = g_strdup_printf("CREATE TABLE IF NOT EXISTS GENRE (%s, %s, UNIQUE(name));",
				"id INTEGER PRIMARY KEY",
				"name VARCHAR(255)");
	if (!exec_sqlite_query(query, cwin, NULL))
		return -1;


	/* Create 'YEAR' table */

	query = g_strdup_printf("CREATE TABLE IF NOT EXISTS YEAR (%s, %s, UNIQUE(year));",
				"id INTEGER PRIMARY KEY",
				"year INT");
	if (!exec_sqlite_query(query, cwin, NULL))
		return -1;

	/* Create 'PLAYLIST_TRACKS' table */

	query = g_strdup_printf("CREATE TABLE IF NOT EXISTS PLAYLIST_TRACKS (%s, %s);",
				"file TEXT",
				"playlist INT");
	if (!exec_sqlite_query(query, cwin, NULL))
		return -1;

	/* Create 'PLAYLIST table */

	query = g_strdup_printf("CREATE TABLE IF NOT EXISTS PLAYLIST (%s, %s, UNIQUE(name));",
				"id INTEGER PRIMARY KEY",
				"name VARCHAR(255)");
	if (!exec_sqlite_query(query, cwin, NULL))
		return -1;

	return 0;
}

gint drop_dbase_schema(struct con_win *cwin)
{
	gint ret = 0;
	gchar *query;

	query = g_strdup_printf("DROP TABLE ALBUM");
	if (!exec_sqlite_query(query, cwin, NULL))
		ret = -1;

	query = g_strdup_printf("DROP TABLE ARTIST");
	if (!exec_sqlite_query(query, cwin, NULL))
		ret = -1;

	query = g_strdup_printf("DROP TABLE GENRE");
	if (!exec_sqlite_query(query, cwin, NULL))
		ret = -1;

	query = g_strdup_printf("DROP TABLE LOCATION");
	if (!exec_sqlite_query(query, cwin, NULL))
		ret = -1;

	query = g_strdup_printf("DROP TABLE TRACK");
	if (!exec_sqlite_query(query, cwin, NULL))
		ret = -1;

	query = g_strdup_printf("DROP TABLE YEAR");
	if (!exec_sqlite_query(query, cwin, NULL))
		ret = -1;

	return ret;
}

void rescan_db(gchar *dir_name, gint no_files, GtkWidget *pbar, gint call_recur, struct con_win *cwin)
{
	static gint files_scanned = 0;
	static gdouble fraction = 0.0;
	GDir *dir;
	const gchar *next_file = NULL;
	gchar *ab_file, txt[64];
	GError *error = NULL;

	/* Reinitialize static variables if called from rescan_library_action */

	if (call_recur) {
		files_scanned = 0;
		fraction = 0.0;
	}

	if (cwin->cstate->stop_scan) return;

	dir = g_dir_open(dir_name, 0, &error);
	if (!dir) {
		g_critical("(%s): Unable to open library : %s", __func__, dir_name);
		return;
	}

	next_file = g_dir_read_name(dir);
	while (next_file) {
		if (cwin->cstate->stop_scan) return;
		ab_file = g_strconcat(dir_name, "/", next_file, NULL);
		if (g_file_test(ab_file, G_FILE_TEST_IS_DIR))
			rescan_db(ab_file, no_files, pbar, 0, cwin);
		else {
			files_scanned++;
			fraction = (gdouble)files_scanned / (gdouble)no_files;
			g_sprintf(txt, "%d / %d", files_scanned, no_files);
			gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(pbar), fraction);
			gtk_progress_bar_set_text(GTK_PROGRESS_BAR(pbar), txt);
			add_entry_db(ab_file, cwin);

		}

		/* Have to give control to GTK periodically ... */

		while(gtk_events_pending())
			gtk_main_iteration();

		g_free(ab_file);
		next_file = g_dir_read_name(dir);
	}
	g_dir_close(dir);
}

void init_library_view(struct con_win *cwin)
{
	gint i = 0;
	gchar *query;
	struct db_result result;
	GtkTreeModel *model, *filter_model;
	const gchar *order_str[] = {
		"LOCATION.name ASC",
		"ARTIST.name ASC, TRACK.title ASC",
		"ALBUM.name ASC, TRACK.title ASC",
		"GENRE.name ASC, TRACK.title ASC",
		"ARTIST.name ASC, ALBUM.name ASC, TRACK.track_no ASC",
		"GENRE.name ASC, ARTIST.name ASC, TRACK.title ASC",
		"GENRE.name ASC, ALBUM.name ASC, TRACK.track_no ASC",
		"GENRE.name ASC, ARTIST.name ASC, ALBUM.name ASC, TRACK.track_no ASC"};

	cwin->cstate->view_change = TRUE;
	filter_model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->library_tree));
	model = gtk_tree_model_filter_get_model(GTK_TREE_MODEL_FILTER(filter_model));
	gtk_tree_store_clear(GTK_TREE_STORE(model));

	/* Query and insert entries */

	query = g_strdup_printf("SELECT TRACK.title,ALBUM.name,ARTIST.name,GENRE.name,\
LOCATION.name,LOCATION.id \
FROM TRACK, ALBUM, ARTIST, GENRE, LOCATION \
WHERE ALBUM.id = TRACK.album AND ARTIST.id = TRACK.artist AND GENRE.id = TRACK.genre AND \
LOCATION.id = TRACK.location \
ORDER BY %s;", order_str[cwin->cpref->cur_library_view]);
	exec_sqlite_query(query, cwin, &result);

	for_each_result_row(result, i) {
		add_entry_library(atoi(result.resultp[i+5]),
				  result.resultp[i+4],
				  result.resultp[i+3],
				  result.resultp[i+2],
				  result.resultp[i+1],
				  result.resultp[i],
				  model,
				  cwin);

		/* Have to give control to GTK periodically ... */
		/* If gtk_main_quit has been called, return - since main loop is no more. */

		while(gtk_events_pending()) {
			if (gtk_main_iteration_do(FALSE)) {
				sqlite3_free_table(result.resultp);
				return;
			}
		}
	}

	sqlite3_free_table(result.resultp);
	cwin->cstate->view_change = FALSE;
}

void library_tree_row_activated_cb(GtkTreeView *library_tree,
				   GtkTreePath *path,
				   GtkTreeViewColumn *column,
				   struct con_win *cwin)
{
	GtkTreeIter iter;
	GtkTreeModel *filter_model;
	gint node_type;

	filter_model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->library_tree));
	gtk_tree_model_get_iter(filter_model, &iter, path);
	gtk_tree_model_get(filter_model, &iter, L_NODE_TYPE, &node_type, -1);

	switch(node_type) {
	case NODE_ARTIST:
	case NODE_ALBUM:
	case NODE_GENRE:
	case NODE_FOLDER:
		if (!gtk_tree_view_row_expanded(GTK_TREE_VIEW(cwin->library_tree), path))
			gtk_tree_view_expand_row(GTK_TREE_VIEW(cwin->library_tree), path, FALSE);
		else
			gtk_tree_view_collapse_row(GTK_TREE_VIEW(cwin->library_tree), path);
		break;
	case NODE_TRACK:
	case NODE_BASENAME:
		add_row_current_playlist(path, filter_model, cwin);
		break;
	default:
		break;
	}
}

gboolean library_tree_right_click_cb(GtkWidget *widget,
				     GdkEventButton *event,
				     struct con_win *cwin)
{
	GtkWidget *popup_menu;
	GtkTreeSelection *selection;
	gboolean ret = FALSE;

	switch(event->button) {
	case 3:
		popup_menu = gtk_ui_manager_get_widget(cwin->library_tree_context_menu, "/popup");
		gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL, NULL, NULL,
			       event->button, event->time);

		selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cwin->library_tree));

		/* If more than one track is selected, don't propagate event */

		if (gtk_tree_selection_count_selected_rows(selection) > 1)
			ret = TRUE;
		else
			ret = FALSE;
		break;
	default:
		ret = FALSE;
		break;
	}

	return ret;
}

gboolean library_page_right_click_cb(GtkWidget *widget,
				     GdkEventButton *event,
				     struct con_win *cwin)
{
	GtkWidget *popup_menu;
	gboolean ret = FALSE;

	switch(event->button) {
	case 3: {
		if ((gtk_notebook_get_current_page(GTK_NOTEBOOK(cwin->browse_mode)) == 0) &&
		    (!cwin->cstate->view_change)) {
			popup_menu = gtk_ui_manager_get_widget(cwin->library_page_context_menu, "/popup");
			gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL, NULL, NULL,
				       event->button, event->time);
		}
		ret = TRUE;
	}
	default:
		ret = FALSE;
		break;
	}
	return ret;
}


/* Callback for DnD signal 'drag-data-get' */

void dnd_library_tree_get(GtkWidget *widget,
			  GdkDragContext *context,
			  GtkSelectionData *data,
			  guint info,
			  guint time,
			  struct con_win *cwin)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model, *filter_model;
	GList *list = NULL, *l;
	GArray *loc_arr;

	switch(info) {
	case TARGET_LOCATION_ID:
		filter_model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->library_tree));
		selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cwin->library_tree));
		list = gtk_tree_selection_get_selected_rows(selection, &model);

		/* No selections */

		if (!list) {
			gtk_selection_data_set(data, data->type, 8, NULL, 0);
			break;
		}

		/* Form an array of location ids */

		loc_arr = g_array_new(TRUE, TRUE, sizeof(gint));
		l = list;

		while(l) {
			get_location_ids(l->data, loc_arr, model, cwin);
			gtk_tree_path_free(l->data);
			l = l->next;
		}

		gtk_selection_data_set(data,
				       data->type,
				       8,
				       (guchar *)GPOINTER_TO_INT(&loc_arr),
				       sizeof(gint));

		/* Cleanup */

		g_list_free(list);

		break;
	default:
		g_warning("(%s): Unknown DND type", __func__);
	}
}

gboolean set_all_visible(GtkTreeModel *model,
			 GtkTreePath *path,
			 GtkTreeIter *iter,
			 gpointer data)
{
	gtk_tree_store_set(GTK_TREE_STORE(model), iter,
			   L_VISIBILE, TRUE, -1);
	return FALSE;
}

void filter_tree_expand_func(GtkTreeView *view,
			     GtkTreePath *path,
			     gpointer data)
{
	struct con_win *cwin = data;
	GtkTreeModel *filter_model;
	GtkTreeIter iter;
	gchar *node_data;
	gint node_type;

	filter_model = gtk_tree_view_get_model(view);
	gtk_tree_model_get_iter(filter_model, &iter, path);
	gtk_tree_model_get(filter_model, &iter, L_NODE_DATA, &node_data, -1);
	gtk_tree_model_get(filter_model, &iter, L_NODE_TYPE, &node_type, -1);

	/* Collapse any non-leaf node that matches the seach entry */

	if (cwin->cstate->filter_entry &&
	    (node_type != NODE_TRACK) &&
	    (node_type != NODE_BASENAME) &&
	    !g_strncasecmp(node_data,
			   cwin->cstate->filter_entry,
			   strlen(cwin->cstate->filter_entry))) {
		gtk_tree_view_collapse_row(view, path);
	}

	g_free(node_data);
}

gboolean filter_tree_func(GtkTreeModel *model,
			  GtkTreePath *path,
			  GtkTreeIter *iter,
			  gpointer data)
{
	struct con_win *cwin = data;
	gchar *node_data = NULL, *t_node_data;
	gboolean visible, t_flag = FALSE;
	GtkTreePath *t_path;
	GtkTreeIter t_iter;

	/* Mark node and its parents visible if search entry matches.
	   If search entry doesn't match, check if _any_ ancestor has
	   been marked as visible and if so, mark current node as visible too. */

	if (cwin->cstate->filter_entry) {
		gtk_tree_model_get(model, iter, L_NODE_DATA, &node_data, -1);
		if (!g_strncasecmp(node_data,
				   cwin->cstate->filter_entry,
				   strlen(cwin->cstate->filter_entry))) {
			gtk_tree_store_set(GTK_TREE_STORE(model), iter,
					   L_VISIBILE, TRUE, -1);
			t_path = gtk_tree_model_get_path(model, iter);
			while (gtk_tree_path_up(t_path)) {
				if (gtk_tree_path_get_depth(t_path) > 0) {
					gtk_tree_model_get_iter(model, &t_iter, t_path);
					gtk_tree_store_set(GTK_TREE_STORE(model), &t_iter,
							   L_VISIBILE, TRUE, -1);
				}
			}
			gtk_tree_path_free(t_path);
		} else {
			t_path = gtk_tree_model_get_path(model, iter);
			while (gtk_tree_path_up(t_path)) {
				if (gtk_tree_path_get_depth(t_path) > 0) {
					gtk_tree_model_get_iter(model, &t_iter, t_path);
					gtk_tree_model_get(model, &t_iter, L_NODE_DATA, &t_node_data, -1);
					gtk_tree_model_get(model, &t_iter, L_VISIBILE, &visible, -1);
					if (visible && !g_strncasecmp(t_node_data,
								      cwin->cstate->filter_entry,
								      strlen(cwin->cstate->filter_entry)))
						t_flag = TRUE;
					g_free(t_node_data);
				}
			}

			if (t_flag)
				gtk_tree_store_set(GTK_TREE_STORE(model), iter,
						   L_VISIBILE, TRUE, -1);
			else
				gtk_tree_store_set(GTK_TREE_STORE(model), iter,
						   L_VISIBILE, FALSE, -1);

			gtk_tree_path_free(t_path);
		}
		g_free(node_data);
	}
	else
		return TRUE;

	return FALSE;
}

void folder_file_library_tree(GtkAction *action, struct con_win *cwin)
{
	free_library_tree_nodes(cwin);
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_FOLDER_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_BASENAME_STR));
	cwin->cpref->cur_library_view = FOLDER_FILE;
	init_library_view(cwin);
}

void artist_track_library_tree(GtkAction *action, struct con_win *cwin)
{
	free_library_tree_nodes(cwin);
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_ARTIST_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_TITLE_STR));
	cwin->cpref->cur_library_view = ARTIST_TRACK;
	init_library_view(cwin);
}

void album_track_library_tree(GtkAction *action, struct con_win *cwin)
{
	free_library_tree_nodes(cwin);
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_ALBUM_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_TITLE_STR));
	cwin->cpref->cur_library_view = ALBUM_TRACK;
	init_library_view(cwin);
}

void genre_track_library_tree(GtkAction *action, struct con_win *cwin)
{
	free_library_tree_nodes(cwin);
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_GENRE_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_TITLE_STR));
	cwin->cpref->cur_library_view = GENRE_TRACK;
	init_library_view(cwin);
}

void artist_album_track_library_tree(GtkAction *action, struct con_win *cwin)
{
	free_library_tree_nodes(cwin);
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_ARTIST_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_ALBUM_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_TITLE_STR));
	cwin->cpref->cur_library_view = ARTIST_ALBUM_TRACK;
	init_library_view(cwin);
}

void genre_album_track_library_tree(GtkAction *action, struct con_win *cwin)
{
	free_library_tree_nodes(cwin);
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_GENRE_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_ALBUM_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_TITLE_STR));
	cwin->cpref->cur_library_view = GENRE_ALBUM_TRACK;
	init_library_view(cwin);
}

void genre_artist_track_library_tree(GtkAction *action, struct con_win *cwin)
{
	free_library_tree_nodes(cwin);
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_GENRE_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_ARTIST_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_TITLE_STR));
	cwin->cpref->cur_library_view = GENRE_ARTIST_TRACK;
	init_library_view(cwin);
}

void genre_artist_album_track_library_tree(GtkAction *action, struct con_win *cwin)
{
	free_library_tree_nodes(cwin);
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_GENRE_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_ARTIST_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_ALBUM_STR));
	cwin->cpref->library_tree_nodes = g_slist_append(cwin->cpref->library_tree_nodes,
							 g_strdup(P_TITLE_STR));
	cwin->cpref->cur_library_view = GENRE_ARTIST_ALBUM_TRACK;
	init_library_view(cwin);
}

void library_tree_play(GtkAction *action, struct con_win *cwin)
{
	GtkTreeModel *model, *filter_model;
	GtkTreeSelection *selection;
	GtkTreePath *path;
	GList *list, *i;

	filter_model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->library_tree));
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cwin->library_tree));
	list = gtk_tree_selection_get_selected_rows(selection, &model);

	if (list) {

		clear_current_playlist(action, cwin);

		/* Add all the rows to the current playlist */

		for (i=list; i != NULL; i = i->next) {
			path = i->data;
			add_row_current_playlist(path, model, cwin);
			gtk_tree_path_free(path);

			/* Have to give control to GTK periodically ... */
			/* If gtk_main_quit has been called, return - since main loop is no more. */

			while(gtk_events_pending())
				if (gtk_main_iteration_do(FALSE))
					return;
		}
		
		g_list_free(list);
	}

	play_first_current_playlist(cwin);
}

void library_tree_enqueue(GtkAction *action, struct con_win *cwin)
{
	GtkTreeModel *model, *filter_model;
	GtkTreeSelection *selection;
	GtkTreePath *path;
	GList *list, *i;

	filter_model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->library_tree));
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cwin->library_tree));
	list = gtk_tree_selection_get_selected_rows(selection, &model);

	if (list) {

		/* Add all the rows to the current playlist */

		for (i=list; i != NULL; i = i->next) {
			path = i->data;
			add_row_current_playlist(path, model, cwin);
			gtk_tree_path_free(path);

			/* Have to give control to GTK periodically ... */
			/* If gtk_main_quit has been called, return - since main loop is no more. */

			while(gtk_events_pending())
				if (gtk_main_iteration_do(FALSE))
					return;
		}
		
		g_list_free(list);
	}
}

/* Get the names of all the playlists stored in the DB.
   Returned NULL terminated array of strings that has to freed by caller. */

gchar** get_playlist_names_db(struct con_win *cwin)
{
	gchar *query;
	struct db_result result;
	gchar **playlists = NULL;
	gint i, j=0;

	query = g_strdup_printf("SELECT NAME FROM PLAYLIST;");
	if (exec_sqlite_query(query, cwin, &result)) {
		if (result.no_rows) {
			playlists = g_malloc0((result.no_rows+1) * sizeof(gchar *));
			for_each_result_row(result, i) {
				playlists[j] = g_strdup(result.resultp[i]);
				j++;
			}
			playlists[j] = NULL;
		}
		sqlite3_free_table(result.resultp);
	}

	return playlists;
}

void new_playlist_db(const gchar *playlist, gint type, struct con_win *cwin)
{
	gint playlist_id = 0;
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (!playlist || !strlen(playlist)) {
		g_warning("(%s): Playlist name is NULL", __func__);
		return;
	}

	model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->current_playlist));
	if (!gtk_tree_model_get_iter_first(model, &iter)) {
		g_warning("(%s): Current playlist is empty", __func__);
		return;
	}

	if ((playlist_id = get_playlist_id(playlist, cwin))) {
		if (overwrite_existing_playlist(playlist, cwin))
			delete_playlist_db((gchar *)playlist, cwin);
		else
			return;
	}

	playlist_id = add_new_playlist_db(playlist, cwin);
	do_save_playlist(playlist_id, type, cwin);
}

void append_playlist_db(const gchar *playlist, gint type, struct con_win *cwin)
{
	gint playlist_id;
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (!playlist || !strlen(playlist)) {
		g_warning("(%s): Playlist name is NULL", __func__);
		return;
	}

	playlist_id = get_playlist_id(playlist, cwin);

	if (!playlist_id) {
		g_warning("(%s): Playlist doesn't exist\n", __func__);
		return;
	}

	model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->current_playlist));
	if (!gtk_tree_model_get_iter_first(model, &iter)) {
		g_warning("(%s): Current playlist is empty", __func__);
		return;
	}

	do_save_playlist(playlist_id, type, cwin);
}

void delete_playlist_db(gchar *playlist, struct con_win *cwin)
{
	gint playlist_id;
	gchar *query;

	if (!playlist || !strlen(playlist)) {
		g_warning("(%s): Playlist name is NULL", __func__);
		return;
	}

	playlist_id = get_playlist_id(playlist, cwin);

	if (!playlist_id) {
		g_warning("(%s): Playlist doesn't exist\n", __func__);
		return;
	}

	query = g_strdup_printf("DELETE FROM PLAYLIST_TRACKS WHERE PLAYLIST=%d;",
				playlist_id);
	exec_sqlite_query(query, cwin, NULL);

	query = g_strdup_printf("DELETE FROM PLAYLIST WHERE ID=%d;",
				playlist_id);
	exec_sqlite_query(query, cwin, NULL);
}

gint get_playlist_id(const gchar *playlist, struct con_win *cwin)
{
	gchar *query;
	gint playlist_id = 0;
	struct db_result result;

	query = g_strdup_printf("SELECT id FROM PLAYLIST WHERE name = '%s'",
				playlist);
	if (exec_sqlite_query(query, cwin, &result)) {
		if (result.no_columns)
			playlist_id = atoi(result.resultp[result.no_columns]);
		sqlite3_free_table(result.resultp);
	}

	return playlist_id;
}

gint get_location_id(const gchar *location, struct con_win *cwin)
{
	gchar *query;
	gint location_id = 0;
	struct db_result result;

	query = g_strdup_printf("SELECT id FROM LOCATION WHERE name = '%s'",
				location);
	if (exec_sqlite_query(query, cwin, &result)) {
		if (result.no_columns)
			location_id = atoi(result.resultp[result.no_columns]);
		sqlite3_free_table(result.resultp);
	}

	return location_id;
}

gboolean exec_sqlite_query(gchar *query, struct con_win *cwin, struct db_result *result)
{
	gchar *err = NULL;
	gboolean ret = FALSE;

	if (!query)
		return FALSE;

	CDEBUG("(%s):%s", __func__, query);

	/* Caller doesn't expect any result */

	if (!result) {
		sqlite3_exec(cwin->cdbase->db, query, NULL, NULL, &err);
		if (err) {
			sqlite3_free(err);
			ret = FALSE;
		}
		else
			ret = TRUE;
	}

	/* Caller expects result */

	else {
		sqlite3_get_table(cwin->cdbase->db, query,
				  &result->resultp,
				  &result->no_rows,
				  &result->no_columns,
				  &err);
		if (err) {
			g_critical("(%s): SQL Err : %s", __func__,  err);
			sqlite3_free(err);
			ret = FALSE;
		}
		else
			ret = TRUE;
	}

	/* Free the query here, don't free in the callsite ! */

	g_free(query);

	return ret;
}
