/*************************************************************************/
/* Copyright (C) 2007,2008 sacamano <m.sacamano@gmail.com>		 */
/* 									 */
/* This program is free software: you can redistribute it and/or modify	 */
/* it under the terms of the GNU General Public License as published by	 */
/* the Free Software Foundation, either version 3 of the License, or	 */
/* (at your option) any later version.					 */
/* 									 */
/* This program is distributed in the hope that it will be useful,	 */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	 */
/* GNU General Public License for more details.				 */
/* 									 */
/* You should have received a copy of the GNU General Public License	 */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*************************************************************************/

#include "consonance.h"
#include <tag_c.h>

gboolean get_wav_info(gchar *file, struct tags *tags)
{
	SNDFILE *sfile = NULL;
	SF_INFO sinfo;

	memset(&sinfo, 0, sizeof(sinfo));

	sfile = sf_open(file, SFM_READ, &sinfo);
	if (!sfile) {
		g_warning("(%s): Unable to open file using sndfile : %s", __func__, file);
		return FALSE;
	}

	tags->artist = g_strdup("");
	tags->album = g_strdup("");
	tags->genre = g_strdup("");
	tags->title = g_path_get_basename(file);
	tags->channels = sinfo.channels;
	tags->samplerate = sinfo.samplerate;
	tags->length = sinfo.frames / sinfo.samplerate;

	if (sf_close(sfile))
		g_warning("(%s): Unable to close file using sndfile : %s", __func__, file);

	return TRUE;
}

static gboolean get_info_taglib(gchar *file, struct tags *tags)
{
	gboolean ret = FALSE;
	TagLib_File *tfile;
	TagLib_Tag *tag;
	const TagLib_AudioProperties *audio_prop;

	tfile = taglib_file_new(file);
	if (!tfile) {
		g_warning("(%s): Unable to open file using taglib : %s", __func__, file);
		ret = FALSE;
		goto exit;
	}

	tag = taglib_file_tag(tfile);
	if (!tag) {
		g_warning("(%s): Unable to locate tag", __func__);
		ret = FALSE;
		goto exit;
	}

	audio_prop = taglib_file_audioproperties(tfile);
	if (!audio_prop) {
		g_warning("(%s): Unable to locate audio properties", __func__);
		ret = FALSE;
		goto exit;
	}

	tags->title = g_strdup(taglib_tag_title(tag));
	tags->artist = g_strdup(taglib_tag_artist(tag));
	tags->album = g_strdup(taglib_tag_album(tag));
	tags->genre = g_strdup(taglib_tag_genre(tag));
	tags->track_no = taglib_tag_track(tag);
	tags->year = taglib_tag_year(tag);
	tags->bitrate = taglib_audioproperties_bitrate(audio_prop);
	tags->length = taglib_audioproperties_length(audio_prop);
	tags->channels = taglib_audioproperties_channels(audio_prop);
	tags->samplerate = taglib_audioproperties_samplerate(audio_prop);
	ret = TRUE;
exit:
	taglib_tag_free_strings();
	taglib_file_free(tfile);

	return ret;
}

gboolean get_mp3_info(gchar *file, struct tags *tags)
{
	return get_info_taglib(file, tags);
}

gboolean get_flac_info(gchar *file, struct tags *tags)
{
	return get_info_taglib(file, tags);
}

gboolean get_ogg_info(gchar *file, struct tags *tags)
{
	return get_info_taglib(file, tags);
}
